@@ -1,5 +1,5 @@
 # engine/url.py
-# Copyright (C) 2005-2017 the SQLAlchemy authors and contributors
+# Copyright (C) 2005-2018 the SQLAlchemy authors and contributors
 # <see AUTHORS file>
 #
 # This module is part of SQLAlchemy and is released under
@@ -54,7 +54,7 @@ class URL(object):
                  host=None, port=None, database=None, query=None):
         self.drivername = drivername
         self.username = username
-        self.password = password
+        self.password_original = password
         self.host = host
         if port is not None:
             self.port = int(port)
@@ -105,6 +105,17 @@ class URL(object):
             self.database == other.database and \
             self.query == other.query
 
+    @property
+    def password(self):
+        if self.password_original is None:
+            return None
+        else:
+            return util.text_type(self.password_original)
+
+    @password.setter
+    def password(self, password):
+        self.password_original = password
+
     def get_backend_name(self):
         if '+' not in self.drivername:
             return self.drivername
@@ -222,7 +233,7 @@ def _parse_rfc1738_args(name):
             query = (
                 len(tokens) > 1 and dict(util.parse_qsl(tokens[1]))) or None
             if util.py2k and query is not None:
-                query = dict((k.encode('ascii'), query[k]) for k in query)
+                query = {k.encode('ascii'): query[k] for k in query}
         else:
             query = None
         components['query'] = query
